import requests
import json

s = requests.Session()


def get_json(url, **kwargs):
    response = s.get(url, timeout=120, **kwargs)
    response.raise_for_status()
    try:
        return response.json()
    except json.JSONDecodeError:
        raise ValueError(f"A request to {url} didn't produce any JSON.")



from pandas import DataFrame
from .request import get_json

metadata_list = ["codigo_sgs", "title", "periodicidade", "unidade_medida"]


def return_search_results_bcb(url):
    json = get_json(url)
    search_results = json["result"]["results"]
    assert search_results, "Nothing was found."
    return DataFrame(search_results).loc[:, metadata_list]


def return_search_results_ipea(url):
    json = get_json(url)
    search_results = json["value"]
    assert search_results, "Nothing was found."
    return DataFrame(search_results)



import re
from .utils import cat, isiterable
from .ipea_metadata_list import ipea_metadata_list
from .dates import today_date, month_to_quarter, check_if_quarters, parse_dates

## IPEA


def ipea_make_select_query(fields):
    # to get the string
    # SERCODIGO,PERNOME,UNINOME,SERNOME,ANOTHERFILTER,ANOTHERFILTER
    # where ANOTHER must be something not alreay selected by default
    defaults = ["SERCODIGO", "SERNOME", "PERNOME", "UNINOME"]
    selected = defaults + [field for field in fields if field not in defaults]
    return f"?$select={','.join(selected)}"


def ipea_make_filter_query(name, fields):
    # to get the string "&$filter=contains(SERNOME,'name')
    # and contains(ANOTHER,'value') and contains(ANOTHER,'value')"
    if any([field not in ipea_metadata_list for field in fields]):
        raise ValueError(f"Can't search for {' or '.join(fields)}. Call ipea.list_fields() if you need help.")
    prefix = "&$filter="
    filter_by_name = f"contains(SERNOME,'{name}')" if name else ""
    filter_by_metadata = ""
    metadata_filters = []
    if fields:
        for metadata, value in fields.items():
            if re.search("(CODIGO|NUMERICA|STATUS)$", metadata):
                if isinstance(value, list):
                    metadata_filters.append("(" + " or ".join([f"{metadata} eq {quote_if_str(item)}" for item in value]) + ")")
                else:
                    metadata_filters.append(f"{metadata} eq {quote_if_str(value)}")
            else:
                metadata_filters.append(f"contains({metadata},'{value}')")
            filter_by_metadata = " and " if filter_by_name else ""
            filter_by_metadata += " and ".join(metadata_filters)
    return f"{prefix}{filter_by_name}{filter_by_metadata}"


def quote_if_str(something):
    return f"'{something}'" if isinstance(something, str) else f"{something}"


## IBGE


def ibge_build_classification_query(classifications=None):
    """
    Auxiliary function to build classifications
    part of the URL.
    """
    if isinstance(classifications, dict):
        s = []
        for classification, category in classifications.items():
            if not category or category == "all":
                s.append(f"{classification}[all]")
            else:
                s.append(f"{classification}[{cat(category, ',')}]")
        return "classificacao=" + "|".join(s)
    elif isinstance(classifications, (int, str)):
        return f"classificacao={classifications}[all]"
    elif isinstance(classifications, list):
        piped_classifications = '|'.join([f"{classification}[all]" for classification in classifications])
        return f"classificacao={piped_classifications}"
    else:
        return ""


def ibge_build_dates_query(start=None, end=None, last_n=None, freq=None):
    """
    Auxiliary function to build the date part
    of the URL.
    """
    start, end = parse_dates(start, end, "ibge")
    if last_n:
        dates = f"/periodos/-{last_n}"
    else:
        if freq == "trimestral":
            if end == today_date().strftime("%Y%m"):
                end = month_to_quarter(today_date()).strftime("%Y%m")
            assert check_if_quarters((start, end)), "Invalid quarter. Choose a number between 1 and 4"
        elif freq == "anual":
            start, end = start[:-2], end[:-2]
        dates = f"/periodos/{start}-{end}"
    return dates


def ibge_build_variables_query(variables):
    if isiterable(variables):
        return f"/variaveis/{cat(variables, '|')}"
    elif variables:
        return f"/variaveis/{variables}"
    else:
        return f"/variaveis"


locations_dict = {
    "N6": "city",
    "N3": "state",
    "N2": "macroregion",
    "N7": "mesoregion",
    "N9": "microregion",
    "N1": "brazil",
}


location_ids = {location: code for code, location in locations_dict.items()}


def ibge_build_location_query(city=None, state=None, macroregion=None, microregion=None, mesoregion=None, brazil=None):
    # note-to-self: http://api.sidra.ibge.gov.br/desctabapi.aspx?c=136
    locations = vars()
    prefix = "&localidades="
    query = []
    if all([code is None for code in locations.values()]):
        return prefix + "BR"
    for location, codes in locations.items():
        if location == "brazil" and codes:
            query.append("BR")
        elif isinstance(codes, list):
            query.append(f"{location_ids[location]}[{cat(codes, ',')}]")
        elif type(codes) == int:
            query.append(f"{location_ids[location]}[{codes}]")
        elif codes:
            query.append(f"{location_ids[location]}[all]")
    return prefix + "|".join(query)

# vi: nowrap



import pandas as pd
from .request import get_json


def parse_bcb_response(url, code, name):
    json = get_json(url)
    assert json, print(f"Request for {code} returned nothing.")
    return json_to_dataframe(json, code, name, "data", "valor", "%d/%m/%Y")


def parse_ipea_response(url, code, name):
    json = get_json(url)["value"]
    assert json, print(f"Request for {code} returned nothing.")
    return json_to_dataframe(json, code, name, "VALDATA", "VALVALOR", "%Y-%m-%dT%H:%M:%S")


def json_to_dataframe(json, code, name, date_key, value_key, date_fmt):
    df = pd.DataFrame(json)
    if date_fmt == "%Y-%m-%dT%H:%M:%S":
        df[date_key] = df[date_key].str[:-6]
    df[date_key] = pd.to_datetime(df[date_key], format=date_fmt)
    df[value_key] = df[value_key].astype('float64')
    df = df.set_index(date_key)
    df.columns = [name if name else code]
    df = df.rename_axis("Date")
    return df


def parse_ibge_response(url):
    json = get_json(url)
    assert len(json) > 1, "This request produced no value."
    return ibge_json_to_dataframe(json)


def ibge_json_to_dataframe(json):
    df = pd.DataFrame(json[1:])
    date_key = json[0]["D2C"]
    df.columns = json[0].values()
    try:
        df[date_key] = pd.to_datetime(df[date_key], format="%Y%m")
    except ValueError:
        df[date_key] = pd.to_datetime(df[date_key], format="%Y")
    df["Valor"] = pd.to_numeric(df["Valor"], errors="coerce")
    df = df.rename(columns={date_key: "Date"})
    df = df.set_index("Date")
    df = df.drop(
        [c for c in df.columns if c.endswith("(Código)")] + ["Mês", "Unidade de Medida", "Brasil"],
        axis="columns",
        errors="ignore"
    )
    return df



ipea_metadata_list = {
    "BASNOME": "Nome da base de dados da série.",
    "FNTNOME": "Nome completo da fonte da série, em português.",
    "FNTSIGLA": "Sigla ou nome abreviado da fonte da série, em português.",
    "FNTURL": "URL para o site da fonte da série.",
    "MULNOME": "Nome do fator multiplicador dos valores da série.",
    "PERNOME": "Nome da periodicidade, em português.",
    "SERATUALIZACAO": "Data da última carga de dados na série.",
    "SERCODIGO": "Código único de identificação da série.",
    "SERCOMENTARIO": "Comentários relativos a série, em português.",
    "SERNOME": "Nome da série, em português.",
    "UNINOME": "Nome da unidade dos valores da série.",
    "SERSTATUS": "Indica se uma série macroeconômica ainda é atualizada. Valores: ‘A’ (Ativa) para séries atualizadas ou ‘I’ (Inativa) para séries que não são atualizadas. As séries regionais ou sociais não possuem este metadado.",
    "TEMCODIGO": "Código de identificação do tema ao qual a série está associada.",
    "PAICODIGO": "Código de identificação país ou região (como América Latina, Zona do Euro, etc.) ao qual a série está associada. Deve ser levado em consideração apenas nas séries macroeconômicas (BASNOME = “Macroeconômico”); atualmente todas as séries regionais e sócias se referem ao Brasil, mesmo esta coluna PAICODIGO é nula ou vazia.",
    "SERNUMERICA": "Série numérica (1), série alfanumérica (0)",
}



import pandas as pd
from .helpers.request import get_json
from .helpers.response import parse_ibge_response
from .helpers.utils import do_search
from .helpers.lists import list_regions
from .helpers.url import (
    ibge_build_dates_query,
    ibge_build_variables_query,
    ibge_build_location_query,
    ibge_build_classification_query,
    locations_dict
)


def get_series(
    code,
    variables=None,
    start=None,
    end=None,
    last_n=None,
    city=None,
    state=None,
    macroregion=None,
    microregion=None,
    mesoregion=None,
    brazil=None,
    classifications=None,
):
    """
    Function to get variables associated with a aggregated variable of IBGE's
    SIDRA database.

    Parameters
    ----------
    code : int
        The code of the aggregated variable.

    variables : int or list of ints, default None
        Which variables to select (if None, return all of them).

    start : int or str, default None
        Initial date, year last.

    end : int or str, default None
        Final date, year last.

    last_n : int or str, default None
        Return only last n observations.

    city : str or int a list of them, default None
        Codes of the cities to be selected.

    state : str or int or a list of them, default None
        Codes of the states to be selected.

    macroregion : str or int or a list of them, default None
        Codes of the macroregions to be selected.

    microregion : str or int or a list of them, default None
        Codes of the microregion to be selected.

    mesoregion : str or int or a list of them, default None
        Codes of the mesoregions to be selected.

    classifications : dict, int, str or list, default None
        { classification : [categories], ... }
        classification1
        "classification1"
        list = [classification1, classification2, ... ]

    Returns
    -------
    A DataFrame with series values and metadata.
    """
    baseurl = f"https://servicodados.ibge.gov.br/api/v3/agregados/{code}"
    # handle dates part of url
    frequency = get_frequency(code)
    dates = ibge_build_dates_query(start, end, last_n, frequency)
    # handle variables, locations and classifications part of url
    variables = ibge_build_variables_query(variables)
    locations = ibge_build_location_query(city, state, macroregion, microregion, mesoregion, brazil)
    classifications = ibge_build_classification_query(classifications)
    url = f"{baseurl}{dates}{variables}?{classifications}{locations}&view=flat"
    return parse_ibge_response(url)

## Get metadata


def get_frequency(aggregate_code):
    return list_periods(aggregate_code).loc["frequency", :].values


def get_metadata(aggregate_code):
    url = f"https://servicodados.ibge.gov.br/api/v3/agregados/{aggregate_code}/metadados"
    json = get_json(url)
    return pd.DataFrame.from_dict(json, orient="index", columns=["values"])

## List Metadata Functions


def search(*search, **searches):
    """
    Function to list all aggregated variables of IBGE.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the aggregates.
    """
    json = get_json("https://servicodados.ibge.gov.br/api/v3/agregados")
    df = pd.io.json.json_normalize(
        json, record_path="agregados", meta=["id", "nome"], meta_prefix="pesquisa_"
    )
    if search or searches:
        return do_search(df, search, searches)
    return df


def list_variables(aggregate_code, *search, **searches):
    """
    Function to list all variables associated with an aggregate of IBGE.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the variables.
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3"
    query = f"/agregados/{aggregate_code}/variaveis/all?localidades=BR"
    url = f"{baseurl}{query}"
    json = get_json(url)
    df = pd.io.json.json_normalize(json).iloc[:, :3]
    if search or searches:
        return do_search(df, search, searches)
    return df


def list_locations(aggregate_code):
    """
    Function to list periods of a given aggregate.

    Returns
    -------
    A DataFrame with the frequency, initial and
    final dates.
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3"
    query = f"/agregados/{aggregate_code}/metadados"
    url = f"{baseurl}{query}"
    codes = get_json(url)["nivelTerritorial"]["Administrativo"]
    df = pd.DataFrame({"codes": codes})
    df["parameters"] = [locations_dict.get(code) for code in df.codes]
    return df


def list_periods(aggregate_code):
    """
    Function to list periods of a given aggregate.

    Returns
    -------
    A DataFrame with the frequency, initial and
    final dates.
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3"
    query = f"/agregados/{aggregate_code}/metadados"
    url = f"{baseurl}{query}"
    json = get_json(url)["periodicidade"]
    df = pd.DataFrame.from_dict(json, orient="index", columns=["value"])
    df.index = ["frequency", "start", "end"]
    return df


def list_classifications(aggregate_code, *search, **searches):
    """
    Function to list all classification of a given aggregate.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with the classifications and their categories.
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3/agregados"
    url = f"{baseurl}/{aggregate_code}/metadados"
    json = get_json(url)
    classifications = json["classificacoes"]
    df = pd.io.json.json_normalize(
        classifications,
        "categorias",
        meta=["id", "nome"],
        meta_prefix="classificacao_"
    )
    if search or searches:
        return do_search(df, search, searches)
    return df


def list_states(*search, **searches):
    """
    Function to list all states and their codes.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the states.
    """
    return list_regions("estados", search, searches)


def list_macroregions(*search, **searches):
    """
    Function to list all macroregions and their codes.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the about the states.
    """
    return list_regions("regioes", search, searches)


def list_cities(*search, **searches):
    """
    Function to list all cities and their codes.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the cities.
    """
    return list_regions("municipios", search, searches)


def list_microregions(*search, **searches):
    """
    Function to list all microregions.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the microregions.
    """
    return list_regions("microrregioes", search, searches)


def list_mesoregions(*search, **searches):
    """
    Function to list all mesoregions and their codes.

    Parameters
    ----------
    search : str
        Strings to search.

    where : str, default "nome"
        Where to search.

    Returns
    -------
    A DataFrame with metadata about the mesoregions.
    """
    return list_regions("mesorregioes", search, searches)

# vi: nowrap



from pandas import concat, DataFrame
from .helpers.dates import parse_dates
from .helpers.utils import return_codes_and_names
from .helpers.response import parse_bcb_response
from .helpers.request import get_json
from .helpers.searching import return_search_results_bcb


def get_serie(code, name=None, start=None, end=None, last_n=None):
    """
    Returns a time series from Time Series Management System (SGS)
    of Brazilian Central Bank (BCB).

    Parameters
    ----------
    code : int or str
    The code of the time series.

    name : str
    The name of the series.

    start : str
    Initial date, year last.

    end : str
    End date, year last.

    last_n : inr
    Ignore other arguments and get the last n observations.

    Returns
    -------
    pandas.DataFrame
    """
    assert isinstance(code, str) or isinstance(code, int), "Not a valid code format."
    baseurl = f"https://api.bcb.gov.br/dados/serie/bcdata.sgs.{code}/dados"
    if last_n:
        url = f"{baseurl}/ultimos/{last_n}?formato=json"
    else:
        start, end = parse_dates(start, end, api="bcb")
        dates = f"&dataInicial={start}" if start else start
        dates += f"&dataFinal={end}" if end else end
        url = f"{baseurl}?format=json{dates}"
    return parse_bcb_response(url, code, name)


def get_series(*codes, start=None, end=None, last_n=None, **kwargs):
    """
    Get multiple series all at once in a single data frame.

    Parameters
    ----------
    codes : dict, str, int
        Dictionary like {"name1": cod1, "name2": cod2} or a bunch of code numbers like cod1, cod2.

    start : str
        Initial date.

    end : str
        End date.

    last_n : int
        Get last_n observations.

    **kwargs
        Passed to pandas.concat.

    Returns
    -------
    A DataFrame with the series.
    """
    assert codes, "You must pass at least one code."
    codes, names = return_codes_and_names(*codes)
    return concat(
        (get_serie(code, name, start, end, last_n)
         for code, name in zip(codes, names)),
        axis="columns",
        sort=True,
        **kwargs
    )


def search(name="", *filters, rows=10, start=1):
    """
    Search for a name in the SGS database.

    Parameters
    ----------
    name : str
        String to search.

    start : int or str
        How many results to show.

    start : int or str
        Start showing from this result.

    Returns
    -------
    A DataFrame with the results.
    """
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"q={name}&rows={rows}&start={start}&sort=score desc"
    filter_params = f"&fq={'+'.join([value for value in filters])}" if filters else ""
    url = f"{baseurl}{params}{filter_params}"
    results = return_search_results_bcb(url)
    return results


def get_metadata(code):
    """
    Get metadata of a series given its code.

    Parameters
    ----------
    code : str
        Code to search.

    Returns
    -------
    A dictionary with the results.
    """
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"fq=codigo_sgs:{code}"
    url = f"{baseurl}{params}"
    results = get_json(url)["result"]["results"]
    return DataFrame.from_dict(results[0], orient="index", columns=["values"])



import re
import pandas as pd


def return_codes_and_names(*args):
    """
    Auxiliary function to return codes and names
    of the series.

    If it finds a dictionary, it will use its keys
    as names and values as codes.

    Otherwise, it will just return the requested
    codes as names, so the column names will be
    the codes themselves.
    """
    for arg in args:
        if isinstance(arg, dict):
            return arg.values(), arg.keys()
    return args, args


def cat(something, sep):
    """
    Auxiliary function to join an iterable if it
    is one, delimited by sep, forcing the join by
    coercing the items to be strings.
    """
    return sep.join(map(str, something)) if isiterable(something) else something


def isiterable(something):
    """
    Auxiliary function to test if something is
    an iterable or not.
    """
    try:
        iter(something)
    except TypeError:
        return False
    return True


def do_search(df, search, searches):
    """
    Helper function to search for regex
    in a given column
    """
    df_cols = df.columns.tolist()
    for field in searches.keys():
        if field not in df_cols:
            raise ValueError("{} is a non-existing column.".format(field))
    regex = build_regex(search)
    name_search = f"{df_cols[1]}.str.contains('{regex}')"
    if searches:
        other_searches = []
        for field, search in searches.items():
            regex = build_regex(search)
            other_searches.append(f"{field}.str.contains('{regex}')")
        other_searches = " and " + " and ".join(other_searches)
        return df.query(name_search + other_searches, engine='python')
    return df.query(name_search, engine='python')


def build_regex(strings):
    """
    Build regex by joining strings by sep
    if it is an iterable but don't do it
    if just a string.
    """
    # (?iu) sets unicode and ignore case flags
    flags = r'(?iu)'
    if isinstance(strings, str):
        regex = f"{flags}{strings}"
    elif isiterable(strings):
        regex = f"{flags}{'|'.join(strings)}"
    return regex


def clean_json(json):
    """
    Helper function to transform JSON into
    a DataFrame and clean its columns names.
    """
    df = pd.io.json.json_normalize(json, sep='_')
    df = df.rename(lambda x: x.replace('.', '_'), axis='columns')
    df = df.rename(lambda x: '_'.join(re.split(r'_', x)[-2:]), axis='columns')
    return df



from pandas import DataFrame
from .request import get_json
from .utils import clean_json, do_search


# IBGE


def list_regions(kind_of_region, search, searches):
    url = f"https://servicodados.ibge.gov.br/api/v1/localidades/{kind_of_region}"
    json = get_json(url)
    df = clean_json(json)
    if search or searches:
        return do_search(df, search, searches)
    return df


# IPEA


def list_metadata(resource_path):
    baseurl = "http://www.ipeadata.gov.br/api/odata4/"
    url = f"{baseurl}{resource_path}"
    json = get_json(url)["value"]
    return DataFrame(json)



from pandas import concat, DataFrame
from .helpers.utils import return_codes_and_names
from .helpers.lists import list_metadata
from .helpers.response import parse_ipea_response
from .helpers.searching import return_search_results_ipea
from .helpers.request import get_json
from .helpers.dates import parse_dates
from .helpers.url import ipea_make_select_query, ipea_make_filter_query
from .helpers.ipea_metadata_list import ipea_metadata_list


def get_serie(code, name=None, start=None, end=None):
    """
    Returns a time series from IPEADATA database.

    Parameters
    ----------
    code : int or str
        The code of the time series.

    name : str
        The name of the series.

    start : str
        Initial date, year last.

    end : str
        End date, year last.

    Returns
    -------
    pandas.DataFrame
    """
    assert isinstance(code, str), "Not a valid code format."
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"ValoresSerie(SERCODIGO='{code}')"
    select = "?$select=VALDATA,VALVALOR"
    start, end = parse_dates(start, end, api="ipeadata")
    dates = date_filter(start, end)
    url = f"{baseurl}{resource_path}{select}{dates}"
    serie = parse_ipea_response(url, code, name)
    return serie


def date_filter(start, end):
    """
    Auxiliary function to return the right query
    to filter dates.
    """
    if start and end:
        data = f"&$filter=VALDATA ge {start} and VALDATA le {end}"
    elif start:
        data = f"&$filter=VALDATA ge {start}"
    elif end:
        data = f"&$filter=VALDATA le {end}"
    else:
        data = ""
    return data


def get_series(*codes, start=None, end=None, **kwargs):
    """
    Get multiple series all at once in a single data frame.

    Parameters
    ----------
    codes : dict, str, int
        Dictionary like {"name1": cod1, "name2": cod2}
        or a bunch of code numbers like cod1, cod2.

    start : str
        Initial date.

    end : str
        End date.

    **kwargs
        Arguments to pandas.concat.

    Returns
    -------
    pandas.DataFrame with the requested series.
    """
    assert codes, "You must pass at least one code to be searched."
    codes, names = return_codes_and_names(*codes)
    return concat(
        (get_serie(code, name, start, end) for code, name in zip(codes, names)),
        axis="columns",
        sort=True,
        **kwargs,
    )


def search(SERNOME="", **fields):
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = "Metadados"
    select_query = ipea_make_select_query(fields)
    filter_query = ipea_make_filter_query(SERNOME, fields)
    url = f"{baseurl}{resource_path}{select_query}{filter_query}"
    results = return_search_results_ipea(url)
    return results


def get_metadata(code):
    """
    Returns metadata of a series specified by the a code.
    """
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"Metadados('{code}')"
    url = f"{baseurl}{resource_path}"
    results = get_json(url)
    return DataFrame.from_dict(results["value"][0], orient="index", columns=["values"])


def list_themes():
    """
    List all themes available in the database.
    """
    return list_metadata("Temas")


def list_countries():
    """
    List all countries available in the database.
    """
    return list_metadata("Paises")


def list_fields():
    """
    Pretty print dictionary of metadata with their description as values
    """
    return DataFrame.from_dict(ipea_metadata_list, orient='index', columns=["Descrição"])
